/**
 * All content on this website (including text, images, source
 * code and any other original works), unless otherwise noted,
 * is licensed under a Creative Commons License.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) Open-Xchange Inc., 2006-2012
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 * @author Malte Timmermann <malte.timmermann@open-xchange.com>
 * @author Ingo Schmidt-Rosbiegal <ingo.schmidt-rosbiegal@open-xchange.com>
 */

define('io.ox/office/framework/app/editapplication',
    ['io.ox/files/api',
     'io.ox/office/tk/utils',
     'io.ox/office/tk/config',
     'io.ox/office/tk/dialogs',
     'io.ox/office/framework/app/baseapplication',
     'io.ox/office/framework/app/rtconnection',
     'io.ox/office/framework/app/extensionregistry',
     'io.ox/office/framework/app/toolbaractions',
     'gettext!io.ox/office/framework'
    ], function (FilesAPI, Utils, Config, Dialogs, BaseApplication, RTConnection, ExtensionRegistry, ToolBarActions, gt) {

    'use strict';

    // private static functions ===============================================

    /**
     * Extracts and validates the actions array from the passed data object of
     * a server response.
     *
     * @param {Object} data
     *  The response data object of the server request.
     *
     * @returns {Array|Undefined}
     *  The actions array, if existing, otherwise undefined.
     */
    function getActionsFromData(data) {

        var // the operations array (may be missing completely)
            actions = Utils.getArrayOption(data, 'actions', []);

        // check that all array elements are objects
        return _(actions).all(_.isObject) ? actions : undefined;
    }

    // class EditApplication ==================================================

    /**
     * The base class for all OX Documents applications allowing to edit a
     * document.
     *
     * Triggers the events supported by the base class BaseApplication, and the
     * following additional events:
     * - 'docs:state': When the state of the application has changed. The event
     *      handlers receive the current application state, as also returned by
     *      the method EditApplication.getState().
     * - 'docs:update': When the application receives an update push message
     *      from the real-time framework. The event handlers receive the
     *      complete data object of the update message.
     *
     * @constructor
     *
     * @extends BaseApplication
     *
     * @param {Function} ModelClass
     *  The constructor function of the document model class. See the
     *  BaseApplication base class constructor for details.
     *
     * @param {Function} ViewClass
     *  The constructor function of the view class. See the BaseApplication
     *  base class constructor for details.
     *
     * @param {Function} ControllerClass
     *  The constructor function of the controller class. See the
     *  BaseApplication base class constructor for details.
     *
     * @param {Object} [appOptions]
     *  A map of static application options, that have been passed to the
     *  static method BaseApplication.createLauncher().
     *
     * @param {Object} [launchOptions]
     *  A map of options passed to the core launcher (the ox.launch() method).
     *  The following options are supported:
     *  @param {String} [launchOptions.action]
     *      Controls how to connect the application to a document file. The
     *      following actions are supported:
     *      - 'load': Tries to load the file described in the launch option
     *          'launchOptions.file'.
     *      - 'new': Tries to create a new document. Optionally, the new file
     *          will be a clone of an existing file described in the launch
     *          option 'launchOptions.templateFile'.
     *      - 'convert': Tries to convert the file described in the launch
     *          option 'launchOptions.templateFile' to a native file format.
     *  @param {Object} [launchOptions.file]
     *      The descriptor of the file to be loaded. Used by the action 'load'
     *      only.
     *  @param {String} [launchOptions.folderId]
     *      The identifier of the folder a new file will be created in. Used by
     *      the actions 'new' and 'convert'.
     *  @param {Object} [launchOptions.templateFile]
     *      The descriptor of an existing file used as template for the new
     *      file. Used by the actions 'new' and 'convert'.
     *  @param {Boolean} [launchOptions.preserveFileName=false]
     *      If set to true, the name of the template file will be preserved
     *      when converting a file to a native file format. Otherwise, a
     *      default file name will be generated. Used by the action 'convert'
     *      only.
     *
     * @param {Object} [options]
     *  A map of options to control the properties of the application. Supports
     *  all properties that are supported by the base class BaseApplication.
     *  Additionally, the following options are supported:
     *  @param {Function} [options.postProcessHandler]
     *      A function that will be called after the document operations have
     *      been downloaded and applied, and that post-processes the document
     *      contents while the progress bar is still visible. Will be called in
     *      the context of this application. Must return a Deferred object that
     *      will be resolved or rejected after the document has been
     *      post-processed, and that may notify the progress of the processing
     *      operation.
     *  @param {Function} [options.optimizeOperationsHandler]
     *      A function that can be used to optimize the operations before they
     *      are sent to the server. Receives an operations array as first
     *      parameter, and must return an operations array. Will be called in
     *      the context of this application instance.
     *  @param {Number} [options.realTimeDelay=700]
     *      The duration in milliseconds the realtime framework will collect
     *      messages before sending them to the server.
     */
    var EditApplication = BaseApplication.extend({ constructor: function (ModelClass, ViewClass, ControllerClass, appOptions, launchOptions, options) {

        var // self reference
            self = this,

            // the document model instance
            model = null,

            // application view: contains panes, tool bars, etc.
            view = null,

            // the unique identifier of this application
            clientId = null,

            // connection to the realtime framework
            rtConnection = null,

            // buffer for actions not yet sent to the server
            actionsBuffer = [],

            // new action collecting multiple operations in an undo group
            pendingAction = null,

            // whether actions are currently sent to the server
            sendingActions = false,

            // post-processing of the document after applying import operations
            postProcessHandler = Utils.getFunctionOption(options, 'postProcessHandler'),

            // callback function that implements merging of actions to reduce data transfer to the server
            optimizeOperationsHandler = Utils.getFunctionOption(options, 'optimizeOperationsHandler'),

            // current application state (see method EditApplication.getState())
            state = null,

            // whether the application is currently connected to the server
            connected = false,

            // cached name of the user currently editing the document
            oldEditUser = null,

            // if any operations where generated by the own document model and sent to the server
            locallyModified = false,

            // if any operations where generated by a remote application
            remotelyModified = false,

            // whether the file has been renamed locally
            locallyRenamed = false,

            // whether application is locked globally (no file permissions)
            locked = false,

            // document file is locked by another user (lock set via OX Drive)
            documentFileIsLocked = false,

            // document file is locked by the user
            documentFileLockedByUser = null,

            // whether application is locked globally (internal unrecoverable error)
            internalError = false,

            // whether the user had edit rights before the offline handler was called
            onlineEditRights = false,

            // whether the connection was at least once offline
            offlineHandlerTriggered = false,

            // the update attributes used by server responses
            updateDataAttributes = ['editUserId', 'editUser', 'hasErrors', 'writeProtected', 'activeClients', 'locked', 'lockedByUser', 'failSafeSaveDone', 'wantsToEditUser', 'wantsToEditUserId'];

        // base constructor ---------------------------------------------------

        BaseApplication.call(this, ModelClass, ViewClass, ControllerClass, importHandler, appOptions, launchOptions, {
            flushHandler: flushDocument
        });

        // private methods ----------------------------------------------------

        /**
         * Extracts the update data from the passed data object of a
         * server response.
         *
         * @param {Object} data
         *  The response data object of a server request.
         *
         * @returns {Object}
         *  An object which contains only the update attributes, if
         *  existing, otherwise undefined.
         */
        function getUpdateAttributesFromData(data) {
            var // the return object
                updateData = {},
                // the found attributes
                attrs = _.intersection(_.keys(data), updateDataAttributes);

            // copy only the known update attributes
            _(attrs).each(function (key) {
                updateData[key] = data[key];
            });

            return updateData;
        }

        /**
         * Recalculates the application state. Triggers a 'docs:state' event,
         * if the application state has been changed.
         */
        function updateState() {

            // calculate the new application state
            var newState = internalError ? 'error' :
                    !connected ? 'offline' :
                    (locked || !model.getEditMode()) ? 'readonly' :
                    self.hasUnsavedChanges() ? 'sending' :
                    locallyModified ? 'ready' :
                    'initial';

            // trigger event if state has been changed
            if (state !== newState) {
                state = newState;
                self.getWindow().nodes.outer.attr('data-app-state', state);
                self.trigger('docs:state', state);
            }
        }

        /**
         * Sets this application into the internal error mode.
         */
        function setInternalError() {
            internalError = true;
            connected = false;
            locked = true;
            model.setEditMode(false);
            updateState();
        }

        /**
         * Shows a read-only alert banner according to the current state of the
         * application. If the document is locked due to missing editing
         * permissions, shows an appropriate alert, otherwise shows the name of
         * the current editing user.
         */
        function showReadOnlyAlert(message) {

            var // alert type
                type = 'info',
                // the alert title
                headline = gt('Read-Only Mode');

            // build message text
            if (!message) {
                if (locked) {
                    if (offlineHandlerTriggered) {
                        // Messages due to offline detection
                        if (connected) {
                            message = gt('Connection to server was lost. Please reopen the document to acquire edit rights.');
                        } else {
                            type = 'warning';
                            message = gt('Connection to server was lost. Switched to read-only mode.');
                        }
                    } else if (documentFileIsLocked) {
                        // Messages due to file locking
                        if (documentFileLockedByUser && documentFileLockedByUser.length > 0) {
                            //#. %1$s is the name of the user who has locked the document
                            //#, c-format
                            message = gt('The document has been locked by %1$s.', _.noI18n(documentFileLockedByUser));
                        } else {
                            message = gt('The document has been locked by another user.');
                        }
                    }  else {
                        // General message due to missing permissions
                        message = gt('You do not have permissions to change this document.');
                    }
                } else if (oldEditUser) {
                    message =
                        //#. %1$s is the name of the user who is currently editing the document
                        //#, c-format
                        gt('%1$s is currently editing this document.', _.noI18n(oldEditUser));
                } else {
                    message = gt('Another user is currently editing this document.');
                }
            }

            // show the alert banner
            view.yell(type, message, headline);
        }

        /**
         * Initialization after construction. Will be called once after
         * construction of the application is finished.
         */
        function initHandler() {

            // get references to MVC instances after construction
            model = self.getModel();
            view = self.getView();

            // calculate initial application state
            updateState();

            // update application state when model changes read-only state
            // (this includes all kinds of internal errors)
            model.on('readonly', updateState);

            // check whether editing is globally disabled for this application type
            if (!ExtensionRegistry.supportsEditMode(self.getDocumentType())) {
                locked = true;
                model.setEditMode(false);
                // TODO: replace with generic message, or get specific message from configuration
                showReadOnlyAlert(gt('Editing on Android not supported yet.'));
            }
        }

        /**
         * Handles a lost network connection. Switches the document model to
         * read-only mode.
         */
        function connectionOfflineHandler() {
            offlineHandlerTriggered = true;
            if (connected && !internalError && !self.isInQuit()) {
                onlineEditRights = model.getEditMode();
                connected = false;
                // Currently we don't have a synchronization process for clients
                // therefore being offline means that we have to switch to
                // read-only mode permanently.
                // TODO: must be changed if synchronization for on/offline is available
                locked = true;
                model.setEditMode(false);
                updateState();
                showReadOnlyAlert();
            }
        }

        /**
         * Handles if we get a network connection again. Sets the internal
         * connected flag to true if our internal error state is not set.
         */
        function connectionOnlineHandler() {
            // RT-Library sometimes calls online although the state has not changed,
            // therefore we also check the offlineHandlerTriggered flag. See
            // #27831 for more information.
            if (offlineHandlerTriggered && !connected && !internalError && !self.isInQuit()) {
                connected = true;
                updateState();
                showReadOnlyAlert();
            }
        }

        /**
         * Handles if a connection has been timed out and was reset. The
         * document is set to read-only and the user has to reopen it to
         * edit it again.
         */
        function connectionResetHandler() {
            if (!self.isLocked() && !self.isInQuit()) {
                setInternalError();
                view.yell('warning', gt('Due to a server connection error the editor switched to read-only mode. Please close and reopen the document.'), gt('Connection Error'));
            }
        }

        /**
         * Handles communication with server failed state.
         */
        function connectionTimeoutHandler() {
            if (!self.isLocked() && !self.isInQuit()) {
                setInternalError();
                view.yell('error', gt('Connection to the server has been lost. Please close and reopen the document.'), gt('Connection Error'));
            }
        }

        /**
         * Handles the situation that we have sent actions to the server but
         * we are not the editor registered at the server. Therefore the server
         * doesn't accept our actions and sends us a hangup notification for
         * reloading. Our document contains changes that are not part of the
         * shared document on the server!
         */
        function connectionHangupHandler() {
            if (!self.isInQuit()) {
                setInternalError();
                view.yell('error', gt('Your document changed on the server. Please close and reopen the document.'), gt('Synchronization Error'));
            }
        }

        /**
         * Handles the situation that the server notifies us that another
         * client wants to receive the edit rights. Therefore we are forced
         * to switch to read-only mode and flush our buffer. If that happend
         * we have to notify the server that the edit rights switch can be
         * accomplished.
         *
         * @param {jQuery.Event} event
         *  The jQuery event object.
         *
         * @param {Object} data
         *  The data sent with the notification.
         */
        function handlePrepareLosingEditRights(event, data) {
            var newEditUserName = Utils.getStringOption(data, 'wantsToEditUser', '');

            if (!self.isInQuit()) {
                model.setEditMode(false);
                updateState();

                showReadOnlyAlert((newEditUserName.length > 0) ?
                    //#. %1$s is the user name who has received the edit rights
                    //#, c-format
                    gt('Transferring edit rights to %1$s. Your latest changes will be saved now.', _.noI18n(newEditUserName)) :
                    gt('Transferring edit rights to another user. Your latest changes will be saved now.')
                );

                self.executeDelayed(function () {
                    sendActions().then(function () {
                        // After sending the actions we can acknowledge that
                        // we are ready to lose our edit rights.
                        rtConnection.canLoseEditRights();
                    });
                });
            }
        }

        /**
         * Handles the situation that this client will receive the
         * edit rights as soon as the transfer from the editor
         * has been completed.
         */
        function handleAcceptedAcquireEditRights() {
            if (!self.isInQuit()) {
                view.yell('info', gt('Edit rights are being transferred to you. Please wait a moment.'));
            }
        }

        /**
         * Handles the situation that this client tries to acquire the
         * edit rights but someone else already did and the process is
         * still in progress.
         */
        function handleDeclinedAcquireEditRights() {
            if (!self.isInQuit()) {
                view.yell('warning', gt('Someone else already wants to receive the edit rights for this document. Please try again later.'));
            }
        }

        /**
         * Updates the edit state of the application according to the passed data of
         * a server notification message.
         *
         * @param {Object} data
         *  The data of a server notification message.
         */
        function applyEditState(data) {

            var // unique identifier of current user with edit rights
                editUserId = Utils.getStringOption(data, 'editUserId', ''),
                // display name of current user with edit rights
                editUser = Utils.getStringOption(data, 'editUser', ''),
                // current state of document edit mode
                oldEditMode = model.getEditMode(),
                // new state of document edit mode
                newEditMode = clientId === editUserId,
                // the file descriptor of this application
                file = self.getFileDescriptor();

            // no state information or internal error may have been set while request was running
            if (internalError || (_.intersection(_.keys(data), updateDataAttributes).length === 0)) {
                return;
            }

            // check if we have to update the files view (fail safe save done on server)
            if (Utils.getBooleanOption(data, 'failSafeSaveDone', false)) {
                FilesAPI.propagate('change', file);
            }

            // internal server error: stop editing completely, show permanent error banner
            if (Utils.getBooleanOption(data, 'hasErrors', false)) {
                setInternalError();
                view.yell('error', gt('Synchronization to the server has been lost.'), gt('Server Error'));
                return;
            }

            if (!locked) {
                // check if file lacks write permissions on server
                if (Utils.getBooleanOption(data, 'writeProtected', false)) {
                    locked = true;
                    model.setEditMode(false);
                    showReadOnlyAlert();
                } else if (Utils.getBooleanOption(data, 'locked', false)) {
                    documentFileIsLocked = true;
                    documentFileLockedByUser = Utils.getStringOption(data, 'lockedByUser', '');
                    locked = true;
                    model.setEditMode(false);
                    showReadOnlyAlert();
                }
            }

            // update edit mode (not if document is locked due to missing write permissions)
            if (!locked) {

                // set edit mode at document model
                model.setEditMode(newEditMode);

                // Show success banner, if edit rights gained after read-only
                // mode, or after connection re-established, but not directly
                // after loading the document (oldEditMode === null).
                if (newEditMode && (oldEditMode === false)) {
                    view.yell('success', gt('You have edit rights.'));
                }

                // Show warning if edit rights have been lost, or are still missing
                // after connection has been re-established, or if the edit user has changed.
                if (!newEditMode && (!connected || oldEditMode || (editUser !== oldEditUser))) {
                    oldEditUser = editUser;
                    showReadOnlyAlert();
                }
            }

            // application is now in connected state
            connected = true;
            updateState();
        }

        /**
         * Updates the file state of the application according to the passed
         * data of a server notification message.
         *
         * @param {Object} data
         *  The data of a server notification message.
         */
        function applyFileState(data) {

            var // current file descriptor
                file = self.getFileDescriptor(),
                // the new file name
                newFileName = Utils.getStringOption(data, 'fileName', ''),
                // the new file version
                newFileVersion = Utils.getStringOption(data, 'fileVersion', ''),
                // the changed components of the file descriptor
                changedFile = {};

            // file name changed after a 'rename' operation
            if ((newFileName.length > 0) && (newFileName !== Utils.getStringOption(file, 'filename'))) {
                changedFile.filename = newFileName;
            }

            // file version changed after remote client closes the modified file
            if ((newFileVersion.length > 0) && (newFileVersion !== Utils.getStringOption(file, 'version'))) {
                changedFile.version = newFileVersion;
            }

            if (!internalError && !_.isEmpty(changedFile)) {
                self.updateFileDescriptor(changedFile);
                self.getController().update();
                updateState();
            }
        }

        /**
         * Applies actions and their operations passed with an 'update' message
         * received from a remote editor of the current document in a
         * background task.
         *
         * @param {Object} data
         *  The data of a server notification message.
         *
         * @return {jQuery.Promise}
         *  The Promise of a Deferred object that will be resolved after all
         *  actions have been applied.
         */
        function applyRemoteActions(data) {

            var // all actions contained in the passed update message data
                actions = getActionsFromData(data);

            if (internalError) {
                return $.Deferred().reject();
            }

            // do nothing, if the update message does not contain any actions
            if (!_.isArray(actions) || (actions.length === 0)) {
                return $.when();
            }

            return model.applyActions(actions, { external: true, async: true, delay: 10 }).done(function () {
                // remember that actions have been received from the server
                remotelyModified = true;
            });
        }

        /**
         * Applies the passed application state and actions passed with an
         * 'update' push message received from the real-time framework. Actions
         * will be applied in a background loop to prevent unresponsive browser
         * and warning messages from the browser. If actions from older
         * 'update' messages are still being applied, the new message data
         * passed to this method will be cached, and will be processed after
         * the background task for the old actions is finished.
         *
         * @param {Object} data
         *  The message data received from 'update' messages sent by the
         *  real-time framework.
         *
         * @param {Object} [options]
         *  A map with options controlling the behavior of this method. The
         *  following options are supported:
         *  @param {Boolean} [options.notify=false]
         *      If set to true, the application will trigger a 'docs:update'
         *      event with the message data, after the actions have been
         *      applied.
         */
        var applyUpdateMessageData = this.createSynchronizedMethod(function (data, options) {

            // apply all actions asynchronously
            return applyRemoteActions(data).done(function () {

                // update application state (read-only mode, etc.)
                applyEditState(data);

                // update application state (fileName, fileVersion)
                applyFileState(data);

                // notify all listeners if specified
                if (Utils.getBooleanOption(options, 'notify', false)) {
                    self.trigger('docs:update', data);
                }
            });
        });

        /**
         * Tries to create a new document in the Files application according to
         * the launch options passed to the constructor.
         *
         * @param {Boolean} convertTemplateFile
         *  Whether to convert the template file to a native file format.
         *
         * @returns {jQuery.Promise}
         *  The Promise of a Deferred object that will be resolved if the new
         *  document has been created successfully. Otherwise, the Promise will
         *  be rejected.
         */
        function createNewDocument(convertTemplateFile) {

            var // parameters passed to the server request
                requestParams = null,

                // file options
                folderId = Utils.getStringOption(launchOptions, 'folderId', ''),
                templateFile = Utils.getObjectOption(launchOptions, 'templateFile'),
                preserveFileName = Utils.getBooleanOption(launchOptions, 'preserveFileName', false);

            // folder identifier must exist in launch options
            if (folderId.length === 0) {
                Utils.error('EditApplication.createNewDocument(): cannot create new document without folder identifier');
                return $.Deferred().reject();
            }

            // create the URL options
            requestParams = {
                action: 'createdefaultdocument',
                folder_id: folderId,
                document_type: self.getDocumentType(),
                initial_filename: /*#. default base file name (without extension) for new OX Documents files (will be extended to e.g. 'unnamed(1).docx') */ gt('unnamed'),
                preserve_filename: preserveFileName,
                convert: convertTemplateFile
            };

            // create from existing file
            if (_.isObject(templateFile)) {
                _(requestParams).extend({
                    template_id: templateFile.id,
                    version: templateFile.version
                });
            }

            // show alert banners after document has been imported successfully
            self.on('docs:import:success', function () {

                var // the message text of the alert banner
                    message = null,
                    // the target folder, may differ from the folder in the launch options
                    targetFolderId = Utils.getStringOption(self.getFileDescriptor(), 'folder_id'),
                    // whether the file has been created in another folder
                    copied = folderId !== targetFolderId;

                if (convertTemplateFile) {
                    message = copied ?
                        //#. %1$s is the new file name
                        //#, c-format
                        gt('Document was converted and stored in your default folder as "%1$s".', _.noI18n(self.getFullFileName())) :
                        //#. %1$s is the new file name
                        //#, c-format
                        gt('Document was converted and stored as "%1$s".', _.noI18n(self.getFullFileName()));
                } else if (copied) {
                    message =
                        //#. %1$s is the file name
                        //#, c-format
                        gt('Document "%1$s" was created in your default folder to allow editing.', _.noI18n(self.getFullFileName()));
                }

                if (message) {
                    view.yell('info', message);
                }
            });

            // send the request
            return self.sendRequest({
                    module: BaseApplication.FILTER_MODULE_NAME,
                    params: requestParams
                })
                .done(function (file) {
                    // creation succeeded: set and propagate file descriptor
                    self.registerFileDescriptor(file);
                })
                .then(undefined, function () {
                    return {
                        cause: convertTemplateFile ? 'convert' : 'create',
                        headline: gt('Server Error'),
                        message: convertTemplateFile ? gt('Could not convert document.') : gt('Could not create document.')
                    };
                })
                .promise();
        }

        /**
         * Loads the document described in the current file descriptor.
         *
         * @param {Object} [point]
         *  The save point, if called from fail-restore.
         *
         * @returns {jQuery.Promise}
         *  The promise of a Deferred object that will be resolved when the
         *  document has been loaded, or rejected when an error has occurred.
         */
        function importDocument(point) {

            var // the application window
                win = self.getWindow(),
                // the footer node in the window blocker element
                footerNode = null,
                // the container element showing the remaining import time
                remainingTimeNode = null,
                // the timer that updates the remaining time label
                remainingTimer = null,
                // whether to show a progress bar (enabled after a short delay)
                showProgressBar = false,
                // current progress
                currentProgress = 0,
                // start time of the import process
                importStartTime = 0,
                // start time for processing the document (operations, formatting)
                processingStartTime = 0,
                // lap time for profiling
                lapTime = 0;

            // updates the progress bar if it is visible
            function updateProgressBar(progress) {
                currentProgress = progress;
                if (showProgressBar) { win.busy(progress); }
            }

            // updates the 'time remaining' label in the busy window
            function updateRemainingTime() {

                var // the current time stamp
                    now = _.now(),
                    // estimated total import time, in 1/1000 seconds
                    totalTime = (currentProgress > 0) ? ((now - importStartTime) / currentProgress) : 0,
                    // estimated remaining import time, in 1/1000 seconds
                    remainingTime = (1 - currentProgress) * totalTime,
                    // estimated remaining import time, rounded to blocks of 5 seconds
                    remainingSeconds = Math.round(remainingTime / 5000) * 5,
                    // estimated remaining import time, rounded to minutes
                    remainingMinutes = Math.round(remainingSeconds / 60);

                if (remainingMinutes >= 2) {
                    remainingTimeNode.text(gt.format(
                        //#. %1$d is the number of minutes remaining to import a document
                        //#, c-format
                        gt.ngettext('%1$d minute remaining', '%1$d minutes remaining', remainingMinutes),
                        _.noI18n(remainingMinutes)
                    ));
                } else if (remainingSeconds > 0) {
                    remainingTimeNode.text(gt.format(
                        //#. %1$d is the number of seconds remaining to import a document
                        //#, c-format
                        gt.ngettext('%1$d second remaining', '%1$d seconds remaining', remainingSeconds),
                        _.noI18n(remainingSeconds)
                    ));
                } else {
                    remainingTimeNode.empty();
                }

                // after at least 10 seconds of processing, show a warning box if remaining time is too high
                if ((now - processingStartTime > 10000) && (remainingTime > 120000)) {
                    footerNode.find('.size-warning-node').show().find('.alert').append(footerNode.find('.cancel-node').show());
                }
            }

            // dumps elapsed time (difference of current time and 'time') to console
            function dumpElapsedTime(msg, count) {
                var now = _.now(),
                    totalTime = now - lapTime,
                    timePerElem = (count > 0) ? Utils.round(totalTime / count, 0.001) : 0;
                Utils.info('EditApplication.importDocument(): ' + msg + ' in ' + totalTime + 'ms' + ((timePerElem > 0) ? (' (' + timePerElem + 'ms per element)') : ''));
                lapTime = now;
            }

            // always load the top-level version of the document when restoring (TODO: really?)
            if (_.isObject(point)) {
                self.updateFileDescriptor({ version: 0 });
            }

            // update debug mode from save point
            view.toggleDebugPane(Utils.getBooleanOption(point, 'debugPane', false));
            view.toggleDebugHighlight(Utils.getBooleanOption(point, 'debugHighlight', false));

            // during opened undo group, collect all operations in the same action
            model.on('undogroup:open', function () {
                pendingAction = createAction();
            });

            // undo group closed: send the pending action to the server
            model.on('undogroup:close', function () {
                registerAction(pendingAction);
                pendingAction = null;
            });

            // collect all operations generated by the document model, send them to the server
            model.on('operations:success', function (event, operations, external) {
                // ignore external operations
                if (!external) {
                    if (_.isObject(pendingAction)) {
                        // if inside an undo group, push all operations into the pending action
                        pendingAction.operations = pendingAction.operations.concat(operations);
                    } else {
                        // otherwise, generate and send a self-contained action
                        registerAction(createAction(operations));
                    }
                }
            });

            // error occurred while applying operations
            model.on('operations:error', function () {
                setInternalError();
                if (self.isImportFinished()) {
                    view.yell('error', gt('An unrecoverable error occurred while modifying the document.'), gt('Internal Error'));
                }
            });

            // disable dropping of images onto the application background area
            self.getWindowNode().on('drop', false);

            // show the progress bar, if import takes more than 1.5 seconds
            self.executeDelayed(function () { showProgressBar = true; }, { delay: 1500 });

            // insert own content into the busy blocker element
            view.enterBusy({
                showFileName: true,
                initHandler: function (header, footer, blocker) {

                    // store footer node for outside use
                    footerNode = footer;

                    // on IE, remove stripes (they are displayed too nervous and flickering)
                    if (_.browser.IE) {
                        blocker.find('.progress-striped').removeClass('progress-striped');
                    }

                    // add a warning banner for large documents that will contain the 'Cancel'
                    // button if visible, and a label for remaining minutes/seconds
                    footer.append(
                        $('<div>').addClass('size-warning-node').append(
                            $('<div>').addClass('alert alert-warning').append(
                                $('<div>').addClass('title').text(gt('Warning')),
                                $('<div>').text(gt('Sorry, your document is very large. It will take some time to be loaded.'))
                            )
                        ).hide(),
                        remainingTimeNode = $('<div>')
                    );
                },
                cancelHandler: function () { self.quit(); }
            });

            // start a timer that updates the label showing the remaining import time
            importStartTime = lapTime = _.now();
            remainingTimer = self.repeatDelayed(updateRemainingTime, { delay: 5000, repeatDelay: 1000 });

            // create the real-time connection to the server, register event listeners
            rtConnection = new RTConnection(self, options);
            rtConnection.on({
                online: connectionOnlineHandler,
                offline: connectionOfflineHandler,
                reset: connectionResetHandler,
                timeout: connectionTimeoutHandler
            });

            // connect to server, receive initial operations of the document
            return rtConnection.connect()
                .progress(function (progress) {
                    updateProgressBar(0.05 * progress); // first 5% for download
                })
                .then(function (data) {
                    // successfully connected to real-time framework, actions downloaded

                    var // the initial action from the server response
                        actions = getActionsFromData(data),
                        // 'update' messages collected during import
                        updateMessages = [],
                        // progress size for applying operations
                        applyProgressSize = _.isFunction(postProcessHandler) ? 0.25 : 0.95,
                        // update data retrieved from connect message
                        updateData = getUpdateAttributesFromData(data);

                    // collects the data of 'update' events in the updateMessages array
                    function collectUpdateEvent(event, data) {
                        updateMessages.push(data);
                    }

                    // connect sends us the first update event together with the action data
                    updateMessages.push(updateData);
                    dumpElapsedTime('document imported');
                    processingStartTime = _.now();

                    // get the own unique client identifier
                    clientId = Utils.getStringOption(data, 'clientId', '');

                    // check error conditions
                    if (Utils.getBooleanOption(data, 'hasErrors', false)) {
                        Utils.error('EditApplication.importDocument(): server side error');
                        return $.Deferred().reject({ cause: 'server' });
                    }
                    if (clientId.length === 0) {
                        Utils.error('EditApplication.importDocument(): missing client identifier');
                        return $.Deferred().reject({ cause: 'noclientid' });
                    }
                    if (!_.isArray(actions) || (actions.length === 0)) {
                        Utils.error('EditApplication.importDocument(): missing actions');
                        return $.Deferred().reject({ cause: 'noactions' });
                    }

                    // immediately register for 'update' messages, collect messages while importing the document
                    rtConnection.on('update', collectUpdateEvent);

                    // apply operations while the entire application pane is detached from the DOM
                    view.detachAppPane();

                    // apply actions at document model asynchronously, update progress bar
                    return model.applyActions(actions, { external: true, async: true })
                        .progress(function (progress) {
                            updateProgressBar(0.05 + progress * applyProgressSize);
                        })
                        .always(function () {
                            // insert the application pane into the DOM, needed by post-processing
                            view.attachAppPane();
                        })
                        .then(function () {
                            // successfully applied all actions

                            var // total number of operations, for logging
                                operationCount = 0,
                                operationStateNumber = 0;

                            if (Config.isDebug()) {
                                _(actions).each(function (action) {
                                    operationCount += Utils.getArrayOption(action, 'operations', []).length;
                                });
                                dumpElapsedTime(operationCount + ' operations applied', operationCount);
                            }

                            operationStateNumber = model.getOperationStateNumber();
                            if (operationStateNumber === -1) {
                                // After successfully applied all actions we must
                                // have a valid operation state number. Otherwise
                                // there is a synchronization error and we have to
                                // stop processing the document further showing
                                // an error message.
                                return $.Deferred().reject({ cause: 'applyactions', headline: gt('Synchronization Error') });
                            }

                            // post-processing of the document contents and formatting
                            return (_.isFunction(postProcessHandler) ? postProcessHandler.call(self) : $.when())
                                .progress(function (progress) {
                                    updateProgressBar(0.3 + progress * 0.7);  // 70% for post-processing
                                })
                                .done(function () {

                                    dumpElapsedTime('postprocessing finished');
                                    updateProgressBar(1);

                                    // register handlers for hangup and switching edit rights notifications
                                    rtConnection.on({
                                        hangup: connectionHangupHandler,
                                        preparelosingeditrights: handlePrepareLosingEditRights,
                                        declinedacquireeditrights: handleDeclinedAcquireEditRights,
                                        acceptedacquireeditrights: handleAcceptedAcquireEditRights
                                    });

                                    // update application state (read-only mode, etc.)
                                    applyEditState(data);

                                    // apply all 'update' messages collected during import (wait for the
                                    // 'docs:import:success' event, otherwise, methods that work with the
                                    // method BaseApplication.isImportFinished() will not run correctly)
                                    self.on('docs:import:success', function () {
                                        _(updateMessages).each(applyUpdateMessageData);
                                        // reconnect 'update' events for direct message processing, forward event
                                        rtConnection.off('update', collectUpdateEvent).on('update', function (event, data) {
                                            applyUpdateMessageData(data, { notify: true });
                                        });
                                    });
                                })
                                .then(undefined, function () {
                                    // failure: post-processing failed
                                    return { cause: 'postprocess' };
                                });
                        }, function () {
                            // failure: applying actions failed
                            return { cause: 'applyactions' };
                        });
                }, function () {
                    // failure: not connected to real-time framework
                    return { cause: 'connection', headline: gt('Server Error') };
                })
                .fail(function () {
                    internalError = true;
                })
                .always(function () {
                    remainingTimer.abort();
                    view.leaveBusy();
                })
                .promise();
        }

        /**
         * Callback executed by the base class BaseApplication to import the
         * document. Creates a new document or converts an existing document,
         * if specified in the launch options.
         *
         * @param {Object} [point]
         *  The save point, if called from fail-restore.
         *
         * @returns {jQuery.Promise}
         *  The promise of a Deferred object that will be resolved when the
         *  document has been loaded, or rejected when an error has occurred.
         */
        function importHandler(point) {

            var // the Deferred object for creation of new document or conversion of existing document
                def = null;

            // create new document in OX Files if required
            switch (Utils.getStringOption(launchOptions, 'action', 'load')) {
            case 'load':
                def = $.when();
                break;
            case 'new':
                def = createNewDocument(false);
                break;
            case 'convert':
                def = createNewDocument(true);
                break;
            default:
                Utils.error('EditApplication.importHandler(): unsupported launch action');
                def = $.Deferred().reject();
            }

            // import the document, if creation/conversion has succeeded
            return def.then(function () { return importDocument(point); }).promise();
        }

        /**
         * Sends all pending actions and flushes the document.
         *
         * @returns {jQuery.Promise}
         *  The Promise of a Deferred object that will be resolved when the
         *  real-time message has been acknowledged.
         */
        function flushDocument() {
            return internalError ? $.Deferred().reject() : sendActions().then(function () {
                rtConnection.sendQuery('flushdocument', 20000);
            });
        }

        /**
         * Creates and returns a new action containing operations and
         * additional data.
         *
         * @param {Array} [operations]
         *  The operations contained by the new action. If omitted, an empty
         *  array will be inserted into the action.
         *
         * @returns {Object}
         *  The new action.
         */
        function createAction(operations) {
            return { operations: operations || [] };
        }

        /**
         * Sends all actions that are currently stored in the actions buffer.
         * If called repeatedly while still sending actions, the new actions
         * buffered in the meantime will be stored in another internal buffer,
         * and will be sent afterwards, until no new actions have been inserted
         * in the actions buffer anymore.
         *
         * @returns {jQuery.Promise}
         *  The Promise of a Deferred object that will be resolved when the
         *  last real-time message containing new actions has been
         *  acknowledged, or rejected if any real-time message has failed.
         */
        var sendActions = (function () {

            var // the Deferred object that will be returned to the caller
                resultDef = null;

            // sends the current actions buffer (calls itself recursively)
            function sendActionsBuffer() {

                // try to reduce the number of actions and operations that need to be sent to the server
                // testing optimizations because of problems with second undo/redo
                if (optimizeOperationsHandler) {
                    actionsBuffer = optimizeOperationsHandler.call(self, actionsBuffer);
                }

                // send all existing actions
                rtConnection.sendActions(actionsBuffer)
                .done(function () {
                    if (internalError) {
                        resultDef.reject();
                    } else if (actionsBuffer.length === 0) {
                        resultDef.resolve();
                    } else {
                        sendActionsBuffer();
                    }
                })
                .fail(function (response) {
                    // TODO: error handling
                    resultDef.reject(response);
                });

                actionsBuffer = [];
            }

            return function () {

                // a real-time message is currently running: return the current
                // result Deferred object that will be resolved after all messages
                if (sendingActions) { return resultDef.promise(); }

                // internal error: immediately return with rejected Deferred
                if (internalError) { return $.Deferred().reject(); }

                // no new actions: return immediately
                if (actionsBuffer.length === 0) { return $.when(); }

                // repare the result Deferred object (will be kept unresolved until
                // all actions have been sent, also including subsequent iterations
                resultDef = $.Deferred().always(function () {
                    resultDef = null;
                    sendingActions = false;
                    updateState();
                });
                sendingActions = true;
                updateState();

                // start sending the actions (will call itself recursively)
                sendActionsBuffer();
                return resultDef.promise();
            };

        }()); // end of local scope for sendActions() method

        /**
         * Registers the passed action to be sent to the server. All registered
         * actions will be collected and sent automatically in a debounced
         * call to the sendActions() method.
         *
         * @param {Object} action
         *  A new action to be sent to the server.
         */
        var registerAction = (function () {

            // direct callback: called every time registerAction() is called, push action to array
            function storeAction(action) {
                if (!internalError && (action.operations.length > 0)) {
                    actionsBuffer.push(action);
                    locallyModified = true;
                    updateState();
                }
            }

            // create and return the debounced registerAction() method
            return self.createDebouncedMethod(storeAction, sendActions);

        }()); // registerAction()

        /**
         * The handler function that will be called when the application is
         * asked to be closed. If the edited document has unsaved changes, a
         * dialog will be shown asking whether to save or drop the changes, or
         * to continue editing.
         *
         * @returns {jQuery.Deferred}
         *  A deferred that will be resolved if the application can be closed
         *  (either if it is unchanged, or the user has chosen to save or lose
         *  the changes), or will be rejected if the application will continue
         *  to run (user has cancelled the dialog, save operation failed).
         */
        function beforeQuitHandler() {

            var // the deferred returned to the framework
                def = $.Deferred();

            // don't call sendActions when we are offline or in an error state
            if ((state !== 'offline') && (state !== 'error')) {
                // send pending actions to the server
                sendActions()
                .done(function () {
                    // still pending actions: ask user whether to close the application
                    if (self.hasUnsavedChanges()) {
                        Dialogs.showYesNoDialog({
                            message: gt('This document contains unsaved changes. Do you really want to close?')
                        }).then(
                            function () { def.resolve(); },
                            function () { def.reject(); }
                        );
                    } else {
                        // all actions saved, close application without dialog
                        def.resolve();
                    }
                })
                .fail(function () {
                    def.resolve();  // reject would keep application alive
                });

                // prevent endless waiting for sendActions to complete
                self.executeDelayed(function () {
                    def.resolve({ cause: 'timeout' });
                }, {delay: 5000});

            } else {
                // in case of offline or error don't wait for the promise
                // to be resolved/rejected. This could mean waiting
                // forever.
                def.resolve(); // reject would keep application alive
            }

            return def.promise();
        }

        /**
         * Called before the application will be really closed.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved or rejected if the application
         *  can be savely closed.
         */
        function quitHandler() {

            var // the result deferred
                def = null,
                // the application window
                win = self.getWindow(),
                // the file descriptor of this application
                file = self.getFileDescriptor();

            // evaluates the result of the 'closedocument' call
            function checkCloseDocumentResult(data) {

                var // the result Deferred object
                    def = $.Deferred(),
                    // the error code in the response
                    errorCode = Utils.getIntegerOption(data, 'errorCode', 0),
                    // cause of a rejected call
                    cause = Utils.getStringOption(data, 'cause', 'unknown'),
                    // possible error text from the server
                    error = Utils.getStringOption(data, 'error', ''),
                    // the message text
                    message = null,
                    // the headline text
                    headline = gt('Server Error'),
                    // file name
                    fullFileName = self.getFullFileName();

                // no error, forward response data to piped callbacks
                if (errorCode === 0) {
                    return data;
                }

                // create error message depending on error code
                switch (errorCode) {
                case 100:
                    message = gt('The document is inconsistent and the server was not able to create a backup file.');
                    break;
                case 101:
                    fullFileName = ExtensionRegistry.createErrorFileName(fullFileName);
                    message =
                        //#. %1$s is the new file name
                        //#, c-format
                        gt('The document is inconsistent and was saved as "%1$s".', _.noI18n(fullFileName));
                    break;
                default:
                    if (cause === 'timeout') {
                        headline = gt('Connection Error');
                        message = gt('Connection to the server has been lost.');
                    } else {
                        message = gt('An unknown error occurred on the server.');
                    }
                }

                // show the alert banner
                internalError = true;
                locked = true;
                updateState();
                if (model && view) {
                    model.setEditMode(false);
                    view.yell('error', message, headline);

                    // redefine application's quit() method to reject the Deferred object
                    self.quit = (function () {
                        var origMethod = self.quit;
                        return function () {
                            def.reject();
                            this.quit = origMethod;
                            return this.quit();
                        };
                    }());

                    // back to idle to allow final copy&paste
                    win.idle();
                    def.always(function () { win.busy(); });
                } else {
                    def.reject();
                }

                if (error.length) {
                    Utils.log(error);
                }

                return def.promise();
            }

            // called in case the 'closedocument' call has been rejected
            function closeDocumentResultFailed(data) {

                var // the cause of the failure
                    cause = Utils.getStringOption(data, 'cause', 'unknown'),
                    // a possible explanation for the error
                    error = Utils.getStringOption(data, 'error', '');

                // simulate error with unknown error code, TODO: evaluate response?
                // return new deferred from UI to give user a chance to see the error
                return checkCloseDocumentResult({ errorCode: -1, cause: cause, error: error });
            }

            function propagateChangedFile() {
                return FilesAPI.propagate('change', file);
            }

            win.busy();

            // Don't call closeDocument when we are offline, in an error state or
            // the document has not been modified.
            if ((state !== 'offline') && (state !== 'error')) {
                // notify server to save/close the document file
                def = rtConnection.closeDocument().then(checkCloseDocumentResult, closeDocumentResultFailed);
            } else {
                // error/offline: it's not safe to make remote calls therefore simply close rtConnection
                def = $.when();
            }

            // disconnect real-time connection after document is closed
            if (rtConnection) {
                def.always(function () {
                    rtConnection.destroy();
                    rtConnection = null;
                });
            }

            if (locallyModified || locallyRenamed || remotelyModified) {
                // propagate changed document file to Files application (also in case of an error)
                def = def.then(propagateChangedFile, propagateChangedFile);
            } else if ((Utils.getStringOption(launchOptions, 'action') === 'new') && !('templateFile' in launchOptions)) {
                // delete new document in Files application if no changes have been made at all
                def = def.then(function () { return FilesAPI.remove([file]); });
            }

            return def;
        }

        // methods ------------------------------------------------------------

        /**
         * Returns whether the edited document is an OpenDocument file.
         */
        this.isODF = function () {
            return this.hasFileDescriptor() && ExtensionRegistry.isOpenDocument(this.getFullFileName());
        };

        /**
         * Returns whether the edited document is a template document file.
         */
        this.isTemplateFile = function () {
            return this.hasFileDescriptor() && ExtensionRegistry.isTemplate(this.getFullFileName());
        };

        /**
         * Returns whether the edited document may contain macro scripts.
         */
        this.isScriptable = function (fileName) {
            return this.hasFileDescriptor() && ExtensionRegistry.isScriptable(this.getFullFileName());
        };

        /**
         * Returns the own client identifier that is globally unique.
         *
         * @returns {String}
         *  The own client identifier.
         */
        this.getClientId = function () {
            return clientId;
        };

        /**
         * Sends an 'addresource' message to register the passed resource
         * identifier, and returns a Deferred object that waits for the server
         * response.
         *
         * @param {String} resourceId
         *  The identifier of the resource to be added.
         *
         * @returns {jQuery.Promise}
         *  The Promise of a Deferred object that will be resolved with the
         *  response data object, after the response of the 'addresource'
         *  message has arrived.
         */
        this.addResourceId = function (resourceId) {
            return rtConnection.addResourceId(resourceId);
        };

        /**
         * Renames the file currently edited by this application.
         *
         * @param {String} shortName
         *  The new short file name (without extension).
         *
         * @returns {jQuery.Promise}
         *  The Promise of a Deferred object that will be resolved when the
         *  file has been renamed successfully; or that will be rejected, if
         *  renaming the file has failed.
         */
        this.rename = function (shortName) {

            var // the result Deferred object
                def = null,
                // the original file name
                oldFileName = this.getFullFileName(),
                // the new file name (trim NPCs and spaces at beginning and end, replace embedded NPCs)
                newFileName = Utils.trimAndCleanString(shortName);

            if (newFileName.length === 0) {
                // empty name (after trimming)
                def = $.Deferred().reject();
            } else if (newFileName === this.getShortFileName()) {
                // name does not change
                def = $.when();
            } else {
                newFileName += '.' + this.getFileExtension();
                // Bug 27724: immediately update file descriptor
                this.updateFileDescriptor({ filename: newFileName });
                def = this.sendFilterRequest({ params: {
                    action: 'renamedocument',
                    filename: newFileName
                } })
                .done(function (data) {
                    locallyRenamed = true;
                    self.updateFileDescriptor({ filename: Utils.getStringOption(data, 'fileName', oldFileName) });
                })
                .fail(function () {
                    // back to old file name, if AJAX request has failed
                    self.updateFileDescriptor({ filename: oldFileName });
                    // TODO: show error banner?
                });
            }

            return def.promise();
        };

        /**
         * Returns whether there are actions left that have not been sent
         * successfully to the server.
         *
         * @attention
         *  Do not rename this method. The OX core uses it to decide whether to
         *  show a warning before the browser refreshes or closes the page.
         *
         * @returns {Boolean}
         *  Whether the application currently caches unsaved actions.
         */
        this.hasUnsavedChanges = function () {
            return sendingActions || (actionsBuffer.length > 0);
        };

        /**
         * Explicitly starts sending all pending actions.
         *
         * @internal
         *  Used for testing purposes only.
         *
         * @returns {jQuery.Promise}
         *  The Promise of a Deferred object that will be resolved after all
         *  pending actions have been saved.
         */
        this.save = function () {
            return sendActions();
        };

        /**
         * Returns the current application state.
         *
         * @returns {String}
         *  A string representing the application state. Will be one of the
         *  following values:
         *  - 'error': internal error, application locked permanently.
         *  - 'offline': network connection missing.
         *  - 'readonly': the document cannot be changed.
         *  - 'initial': the document has not been changed yet.
         *  - 'ready': all local changes have been sent to the server.
         *  - 'sending': local changes are currently being sent to the server.
         */
        this.getState = function () {
            return state;
        };

        /**
         * Returns whether the application is locked for editing. Reasons for
         * the lock may be missing edit permissions for the document file, or
         * an unrecoverable internal application error.
         *
         * @returns {Boolean}
         *  Whether editing the document is locked.
         */
        this.isLocked = function () {
            return internalError || locked;
        };

        /**
         * Tries to acquire edit rights for the document. Edit rights will be
         * updated automatically while synchronizing operations with the
         * server. Multiple calls of this function will be throttled with a
         * delay of 3 seconds (first call will be executed immediately).
         */
        this.acquireEditRights = _.throttle(function () {
            if (!self.isLocked() && !model.getEditMode()) {
                rtConnection.acquireEditRights();
            }
        }, 3000);

        /**
         * Launches the appropriate OX Document edit applications with
         * current document. The old application instance will be closed.
         */
        this.reloadDocument = function () {
            var // the file descriptor of the current document
                file = this.getFileDescriptor(),
                // full filename
                fullFileName = this.getFullFileName(),
                // the configuration of the file extension of the current document
                extensionSettings = ExtensionRegistry.getExtensionSettings(fullFileName),
                // the edit application module identifier
                editModule = Utils.getStringOption(extensionSettings, 'module', ''),
                // quit deferred
                quitDefer = $.Deferred();

            if (editModule.length > 0) {

                // close this application
                _.defer(function () { self.quit().then(function () { quitDefer.resolve(); }); });

                // Wait for the quit of the current instance and
                // lanuch the correct edit application. Otherwise the
                // launcher will detect that an instance with the current
                // document is open and just doesn't do anything!
                quitDefer.then(function () {
                    if (ExtensionRegistry.isNative(fullFileName)) {
                        _.delay(function () { ox.launch(editModule + '/main', { action: 'load', file: file }); }, 1000);
                    } else {
                        Utils.error('EditApplication.reloadDocument(): unknown document type');
                        return;
                    }

                });
            }
        };

        /**
         * Rejects an attempt to edit the document (e.g. due to received
         * keyboard events in read-only mode). The application will show an
         * appropriate warning alert banner to the user, and will update the
         * internal state according to the cause of the rejection.
         *
         * @param {String} [cause='edit']
         *  The cause of the rejection. If omitted, a simple edit attempt has
         *  been rejected, without any further influence on the application
         *  state. If set to 'image', inserting an image into the document has
         *  failed. If set to 'siri', a Siri event has been captured (the
         *  application goes into the internal error mode and will be locked).
         *
         * @returns {EditApplication}
         *  A reference to this instance.
         */
        this.rejectEditAttempt = function (cause) {

            switch (cause) {
            case 'image':
                view.yell('warning', gt('The image could not be inserted.'));
                break;

            case 'siri':
                setInternalError();
                showReadOnlyAlert(gt('Siri is not supported as input method. Please do not use to input text.'));
                break;

            default:
                showReadOnlyAlert();
            }

            return this;
        };

        this.triggerDebugAction = Config.isDebug() ? function (action) {

            switch (action) {
            case 'offline':
                rtConnection.trigger('offline');
                break;
            case 'online':
                rtConnection.trigger('online');
                break;
            case 'reset':
                rtConnection.trigger('reset');
                break;
            case 'timeout':
                rtConnection.trigger('timeout');
                break;
            case 'hangup':
                rtConnection.trigger('hangup');
                break;
            case 'invalid-op':
                model.applyOperations({ name: '_invalid_operation_' });
                break;
            case 'quit-unsaved':
                _.defer(function () {
                    var originalMethod = self.hasUnsavedChanges;
                    self.hasUnsavedChanges = function () { return true; };
                    self.quit().fail(function () { self.hasUnsavedChanges = originalMethod; });
                });
                break;
            case 'quit-error':
                rtConnection.closeDocument = function () {
                    var def = $.Deferred();
                    _.delay(function () { def.reject(); }, 2000);
                    return def;
                };
                _.defer(function () { self.quit(); });
                break;
            }

        } : $.noop;

        // initialization -----------------------------------------------------

        // initialization after construction
        this.on('docs:init', initHandler);

        // set handlers (fail-save handler returns data needed when restoring
        // the application after browser refresh)
        this.registerBeforeQuitHandler(beforeQuitHandler)
            .registerQuitHandler(quitHandler)
            .registerFailSaveHandler(function () {
                return { debugPane: view.isDebugPaneVisible(), debugHighlight: view.isDebugHighlight() };
            });

    }}); // class EditApplication

    // static methods ---------------------------------------------------------

    EditApplication.createLauncher = function (moduleName, ApplicationClass, appOptions) {

        // configure the global window tool bar
        if (Utils.getBooleanOption(appOptions, 'search', false)) {
            ToolBarActions.createSearchIcon(moduleName);
        }
        ToolBarActions.createDownloadIcon(moduleName);
        ToolBarActions.createPrintIcon(moduleName);
        ToolBarActions.createMailIcon(moduleName);

        // create and return the application launcher
        return BaseApplication.createLauncher(moduleName, ApplicationClass, appOptions);
    };

    // exports ================================================================

    return EditApplication;

});
